/* FILE: bitmap.cc                 -*-Mode: c++-*-
 *
 * OommfBitmap C++ class code
 *
 */

#include "bitmap.h"

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

#define CODECHECKS

//////////////////////////////////////////////////////////////////////////
// Drawing parameters and OommfPolygon helper functions
static REAL8m ArrowLength(0.90);
static REAL8m BitmapLineWidth(0.08);
static REAL8m BitmapArrowHeadRatio(0.5);
static REAL8m InOutTipRadius(0.075);
static REAL8m DotRadius(0.025);
static REAL8m CrossLineWidth(0.025);
static REAL8m MinCrossDisplayMag(22.);

static void MakeStandardDiamond(OommfPolygon& diamond)
{
  diamond.Clear();
  REAL8m radius=InOutTipRadius*sqrt(2.);
  diamond.AddVertex(PlanePoint(-radius,0.));
  diamond.AddVertex(PlanePoint(0.,-radius));
  diamond.AddVertex(PlanePoint(radius,0.));
  diamond.AddVertex(PlanePoint(0.,radius));
}

static void MakeStandardCross(OommfPolygon& cross)
{
  cross.Clear();
  REAL8m inner=CrossLineWidth/2.;
  REAL8m outer=InOutTipRadius*sqrt(2.)-inner;
  cross.AddVertex(PlanePoint(outer,inner));
  cross.AddVertex(PlanePoint(inner,inner));
  cross.AddVertex(PlanePoint(inner,outer));
  cross.AddVertex(PlanePoint(-inner,outer));
  cross.AddVertex(PlanePoint(-inner,inner));
  cross.AddVertex(PlanePoint(-outer,inner));
  cross.AddVertex(PlanePoint(-outer,-inner));
  cross.AddVertex(PlanePoint(-inner,-inner));
  cross.AddVertex(PlanePoint(-inner,-outer));
  cross.AddVertex(PlanePoint(inner,-outer));
  cross.AddVertex(PlanePoint(inner,-inner));
  cross.AddVertex(PlanePoint(outer,-inner));
}

static void MakeStandardArrow(OommfPolygon& arrow)
{
  arrow.Clear();

  arrow.AddVertex(PlanePoint(-ArrowLength/2.,BitmapLineWidth/2.));
  arrow.AddVertex(PlanePoint(ArrowLength/2.-BitmapArrowHeadRatio*0.8,
                             BitmapLineWidth/2.));
  arrow.AddVertex(PlanePoint(ArrowLength/2.-BitmapArrowHeadRatio,
                             BitmapArrowHeadRatio*0.4));

  arrow.AddVertex(PlanePoint(ArrowLength/2.,BitmapLineWidth/4.));
  arrow.AddVertex(PlanePoint(ArrowLength/2.,-BitmapLineWidth/4.));

  arrow.AddVertex(PlanePoint(ArrowLength/2.-BitmapArrowHeadRatio,
                             -BitmapArrowHeadRatio*0.4));
  arrow.AddVertex(PlanePoint(ArrowLength/2.-BitmapArrowHeadRatio*0.8,
                             -BitmapLineWidth/2.));
  arrow.AddVertex(PlanePoint(-ArrowLength/2.,-BitmapLineWidth/2.));

}

//////////////////////////////////////////////////////////////////////////
// OOMMF Polygon class
const ClassDoc OommfPolygon::class_doc("OommfPolygon",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","31-Oct-1998");

OommfPolygon::OommfPolygon(const OommfPolygon& poly)
{
  vertex=poly.vertex;
}

OommfPolygon& OommfPolygon::operator=(OommfPolygon& poly)
{
#define MEMBERNAME "operator=(OommfPolygon&)"
  vertex=poly.vertex;
  return *this;
#undef MEMBERNAME
}

void OommfPolygon::Clear()
{
  vertex.Clear();  // Note: Does not release memory
}

int OommfPolygon::GetVertexCount()
{
  return vertex.GetSize();
}

void OommfPolygon::AddVertex(const PlanePoint& pt)
{ // Adds a new vertex to polygon
  vertex.Append(pt);
}

void OommfPolygon::GetBounds(REAL8m& xmin,REAL8m& ymin,
               REAL8m& xmax,REAL8m& ymax) const
{ // Returns 0,0,0,0 if polygon is empty
  Nb_List_Index<PlanePoint> key;
  const PlanePoint* pt=vertex.GetFirst(key);
  if(pt==NULL) {
    xmin=ymin=xmax=ymax=0.0;
    return;
  }
  xmin=xmax=pt->x;
  ymin=ymax=pt->y;
  while((pt=vertex.GetNext(key))!=NULL) {
    REAL8m x=pt->x;
    if(xmin>x) xmin=x;
    if(xmax<x) xmax=x;
    REAL8m y=pt->y;
    if(ymin>y) ymin=y;
    if(ymax<y) ymax=y;
  }
}

// CrossSegment returns 1 if the ray from point "Q" in
// direction (1,eps) (where eps is an arbitrarily small positive
// value) crosses the line segment between points P1 and P2.
// The return value is -1 if the distance from "Q" to the line
// segment is approx < REAL4_EPSILON.  Otherwise the return value
// is 0. See page 11-12 (7-Sept-95) of MJD's micromagnetic NIST lab
// notebook.  This routine is used by OommfPolygon::IsInside().
// (Note: The distance comparison is only "approx < REAL4_EPSILON",
// because just to simplify the code the distance is taken in some
// cases wrt the L_\infty metric.  This could be cleaned up if
// necessary.)
static int CrossSegment(const PlanePoint& Q,
                        const PlanePoint& P1,const PlanePoint& P2)
{
  REAL8 delta;  // Discriminant for 2x2 inversion matrix
  delta=P2.y-P1.y;
  if(fabs(delta)<REAL8_EPSILON) {
    // Ray and line are parallel.
    // NOTE: Do _not_ confuse REAL8_EPSILON with eps is preceding
    //       discussion.
    if(fabs(Q.y-P1.y)<REAL4_EPSILON &&
       Q.x>OC_MIN(P1.x,P2.x)-REAL4_EPSILON &&
       Q.x<OC_MAX(P1.x,P2.x)+REAL4_EPSILON)
      return -1;
    return 0;
  }
  REAL8 delta_inv=1./delta;

  REAL8 s=(P1.x-Q.x)+(Q.y-P1.y)*(P2.x-P1.x)*delta_inv;
  /// s is the distance from Q to line through P1 & P2

  REAL8 t=(Q.y-P1.y)*delta_inv;
  /// t is the location of the intersection point on P1+P2,
  /// using convex coordinate representation.

  if(fabs(s)<REAL4_EPSILON &&
     t>(-REAL4_EPSILON) && t<(1+REAL4_EPSILON)) return -1; // Very close!

  // Make sure intersection is on positive side of ray
  if(s<0) return 0;  // Negative side

  // Is intersection inside line segment proper?
  if(delta>0) {
    if(t<0. || t>=1.) return 0;
  } else {
    if(t<=0. || t>1.) return 0;
  }

  // Else we have a valid intersection
  return 1;
}

// OommfPolygon::IsInside returns 1 if "pt" is inside the polygon,
// 0 otherwise.  The technique is to draw a ray from "pt" in
// direction (1,eps) (where eps is an arbitrarily small positive
// value) and count how many polygon edges get crossed.  An even
// count occurs for points outside the polygon, odd for inside.
// The eps shift is used to avoid vertices.  See also page 11-12
// (7-Sept-95) of MJD's micromagnetic NIST lab notebook.
// Additionally (2-Nov-1998), a check is made to see if Q is very
// near (REAL4_EPSILON) the boundary of the polygon.  If so, then
// it is included.  (See item (4) on page 12 of lab notebook.)
int OommfPolygon::IsInside(const PlanePoint& pt) const
{
  const PlanePoint *pt0=NULL,*pta=NULL,*ptb=NULL;
  Nb_List_Index<PlanePoint> key;
  if((pt0=vertex.GetFirst(key))==NULL
     || (ptb=vertex.GetNext(key))==NULL)
    return 0; // Not enough points in polygon
  pta=pt0;
  int intersections,crosscode;
  intersections=0;
  do {
    crosscode=CrossSegment(pt,*pta,*ptb);
    if(crosscode<0) return 1; // pt is on or nearly on this edge
    intersections+=crosscode;
    pta=ptb;
  } while((ptb=vertex.GetNext(key))!=NULL);

  crosscode=CrossSegment(pt,*pta,*pt0);  // Close polygon
  if(crosscode<0) return 1; // pt is on or nearly on this edge
  intersections+=crosscode;
  return intersections%2;
}

// Polygon coordinate transformation routines (Scale, Shift and Rotate)
void OommfPolygon::Scale(REAL8m xscale,REAL8m yscale)
{
  Nb_List_Index<PlanePoint> key;
  for(PlanePoint* pt=vertex.GetFirst(key); pt!=NULL;
      pt=vertex.GetNext(key)) {
    pt->x*=xscale; pt->y*=yscale;
  }
}

void OommfPolygon::Shift(REAL8m xoffset,REAL8m yoffset)
{
  Nb_List_Index<PlanePoint> key;
  for(PlanePoint* pt=vertex.GetFirst(key) ;pt!=NULL;
      pt=vertex.GetNext(key)) {
    pt->x+=xoffset; pt->y+=yoffset;
  }
}

void OommfPolygon::Rotate(REAL8m angcos,REAL8m angsin)
{
  REAL8m tempx,tempy;
  Nb_List_Index<PlanePoint> key;
  for(PlanePoint* pt=vertex.GetFirst(key); pt!=NULL;
      pt=vertex.GetNext(key)) {
    tempx=pt->x; tempy=pt->y;
    pt->x=angcos*tempx-angsin*tempy;
    pt->y=angsin*tempx+angcos*tempy;

  }
}

// Rotates and rescales bitmap as necessary to shift vertex[0]
// to the nearest integral coordinate.  This may make some bitmaps
// prettier.
void OommfPolygon::AdjustFirstToInteger(void)
{
  PlanePoint newpt;
  REAL8 magold,magnew,angcos,angsin,scale;
  Nb_List_Index<PlanePoint> key;
  PlanePoint* pta=vertex.GetFirst(key);
  if(pta==NULL) return; // Empty polygon; do nothing
  magold=pta->Mag();
  if(magold<REAL8_EPSILON) return; // pta is too small to do anything with.
  newpt.x=OC_ROUND(pta->x);
  newpt.y=OC_ROUND(pta->y);
  magnew=newpt.Mag();
  scale=magnew/magold;
  angcos=(pta->x*newpt.x+pta->y*newpt.y)/(magold*magnew);
  angsin=sqrt(1-angcos*angcos);
  if(pta->x*newpt.y<pta->y*newpt.x) angsin*=-1;
  this->Rotate(angcos,angsin);
  this->Scale(scale,scale);
}

// Simply adjusts each vertex individually to nearest integer
void OommfPolygon::AdjustAllToIntegers()
{
  Nb_List_Index<PlanePoint> key;
  for(PlanePoint* pt=vertex.GetFirst(key); pt!=NULL;
      pt=vertex.GetNext(key)) {
    pt->x=OC_ROUND(pt->x);
    pt->y=OC_ROUND(pt->y);
  }
}

void OommfPolygon::DumpVertices(FILE* fout)
{ // Debugging tool
  Nb_List_Index<PlanePoint> key;
  for(PlanePoint* pt=vertex.GetFirst(key); pt!=NULL;
      pt=vertex.GetNext(key)) {
    fprintf(fout,"(%g,%g) ",pt->x,pt->y);
  }
  fprintf(fout,"\n");
}

//////////////////////////////////////////////////////////////////////////
// OOMMF Bitmap class
const ClassDoc OommfBitmap::class_doc("OommfBitmap",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","23-Oct-1998");

void OommfBitmap::Free()
{
  if(bitmap==NULL) return;
  delete[] bitmap;
  bitmap=NULL;
}

void OommfBitmap::Alloc(int new_xmin,int new_ymin,
			int new_xmax,int new_ymax)
{
#define MEMBERNAME "Alloc(int,int,int,int,int)"
  Free();  // Release previous bitmap, if any
  xmin=new_xmin; xmax=new_xmax;
  ymin=new_ymin; ymax=new_ymax;
  xsize=xmax-xmin+1;
  ysize=ymax-ymin+1;
  if(xsize<1 || ysize<1 )
    FatalError(-1,STDDOC,"Illegal bitmap size request: %d x %d",xsize,ysize);
  totalsize=xsize*ysize;
  if(totalsize<xsize || totalsize<ysize)
    FatalError(-1,STDDOC,"Integer overflow detecting allocating bitmap"
	       " of size: %d x %d",xsize,ysize);
  bitmap=new OommfPackedRGB[totalsize];
  if(bitmap==0)
    FatalError(-1,STDDOC,"Insufficient memory for allocating bitmap"
	       " of size: %d x %d",xsize,ysize);
#undef MEMBERNAME
}

OommfBitmap::OommfBitmap() {
  bitmap=NULL;
  MakeStandardArrow(default_arrow);
  MakeStandardDiamond(default_diamond);
  MakeStandardCross(default_cross);
  Setup(0,0,1,1);
}

void OommfBitmap::Setup(int new_xmin,int new_ymin,
                        int new_xmax,int new_ymax)
{
  Alloc(new_xmin,new_ymin,new_xmax,new_ymax);
}

void OommfBitmap::SetPixel(int x,int y,OommfPackedRGB color)
{
#define MEMBERNAME "SetPixel(int,int,OommfPackedRGB)"
  int offset= (y-ymin)*xsize + (x-xmin);
#ifdef CODECHECKS
  if(offset<0 || offset>=totalsize)
    FatalError(-1,STDDOC,"Out-of-bounds bitmap write attempted: "
	       "(%d,%d) (should be < (%d,%d) )",x,y,xsize,ysize);
#endif
  bitmap[offset]=color;
#undef MEMBERNAME
}

OommfPackedRGB OommfBitmap::GetPixel(int x,int y) const
{
#define MEMBERNAME "GetPixel(int,int)"
  int offset= (y-ymin)*xsize + (x-xmin);
#ifdef CODECHECKS
  if(offset<0 || offset>=totalsize)
    FatalError(-1,STDDOC,"Out-of-bounds bitmap read attempted: "
	       "(%d,%d) (should be < (%d,%d) )",x,y,xsize,ysize);
#endif
  return bitmap[offset];
#undef MEMBERNAME
}

void OommfBitmap::Erase(OommfPackedRGB color)
{
  for(int i=0;i<totalsize;i++) bitmap[i]=color;
}

void OommfBitmap::DrawFilledRectangle(REAL8m x1,REAL8m y1,
                                      REAL8m x2,REAL8m y2,
                                      OommfPackedRGB color)
{ // (x1,y1) and (x2,y2) are opposing vertices, and are *both*
  // included in the rectangle.  Note that this differs from the
  // Tk canvas widget specification, in which only the upper and left
  // edges are included (the lower and right are not).
  int ix1,ix2;
  if(x2<x1) { ix1=int(floor(x2)); ix2=int(ceil(x1)); }
  else      { ix1=int(floor(x1)); ix2=int(ceil(x2)); }
  int iy1,iy2;
  if(y2<y1) { iy1=int(floor(y2)); iy2=int(ceil(y1)); }
  else      { iy1=int(floor(y1)); iy2=int(ceil(y2)); }
  if(ix1<xmin) ix1=xmin;
  if(ix2>xmax) ix2=xmax;
  if(iy1<ymin) iy1=ymin;
  if(iy2>ymax) iy2=ymax;
  for(int ix=ix1;ix<=ix2;ix++) for(int iy=iy1;iy<=iy2;iy++) {
    SetPixel(ix,iy,color);
  }
}

void OommfBitmap::DrawRectangleFrame(REAL8m x1,REAL8m y1,
				     REAL8m x2,REAL8m y2,
				     REAL8m borderwidth,
				     OommfPackedRGB color)
{ // A frame is drawn outside the rectangle specified by opposing
  // vertices (x1,y1) and (x2,y2).  The line width of the frame
  // is specified by 'borderwidth', and the color by 'color'.
  if(x1>x2) { REAL8m temp=x1; x2=x1; x1=temp; }
  if(y1>y2) { REAL8m temp=y1; y2=y1; y1=temp; }
  int ix1=int(OC_ROUND(x1))-1;   // Make sure we don't overlap inside
  int ix2=int(OC_ROUND(x2))+1;   // rectangle
  int iy1=int(OC_ROUND(y1))-1;
  int iy2=int(OC_ROUND(y2))+1;
  int iwidth=int(OC_ROUND(borderwidth));

  int ixa,iya,ixb,iyb;  // Working vertices
  int ix,iy;            // Pixel points

  // Top line
  ixa=ix1-iwidth;  ixb=ix2+iwidth;
  iya=iy1-iwidth;  iyb=iy1;
  for(ix=ixa;ix<=ixb;ix++) for(iy=iya;iy<=iyb;iy++) {
    SetPixel(ix,iy,color);
  }

  // Bottom line
  iya=iy2;         iyb=iy2+iwidth;
  for(ix=ixa;ix<=ixb;ix++) for(iy=iya;iy<=iyb;iy++) {
    SetPixel(ix,iy,color);
  }

  // Left edge
  ixa=ix1-iwidth;  ixb=ix1;
  iya=iy1;         iyb=iy2;
  for(ix=ixa;ix<=ixb;ix++) for(iy=iya;iy<=iyb;iy++) {
    SetPixel(ix,iy,color);
  }

  // Right edge
  ixa=ix2;         ixb=ix2+iwidth;
  for(ix=ixa;ix<=ixb;ix++) for(iy=iya;iy<=iyb;iy++) {
    SetPixel(ix,iy,color);
  }

}

void OommfBitmap::DrawFilledPoly(const OommfPolygon& poly,
				 OommfPackedRGB color,BOOL antialias)
{
  REAL8m raxmin,raymin,raxmax,raymax;
  poly.GetBounds(raxmin,raymin,raxmax,raymax);
  int axmin=int(floor(raxmin));  if(axmin<xmin) axmin=xmin;
  int aymin=int(floor(raymin));  if(aymin<ymin) aymin=ymin;
  int axmax=int(floor(raxmax));  if(axmax>xmax) axmax=xmax;
  int aymax=int(floor(raymax));  if(aymax>ymax) aymax=ymax;
  if(antialias) {
#   define AASIZE 4   // This is subsampling array edge count.
    OommfPackedRGB pixcolor;
    /// 2 is definitely not enough, 3 may be okay, 4 is pretty good.
    for(int x=axmin;x<=axmax;x++) for(int y=aymin;y<=aymax;y++) {
      int incount=0;
      for(REAL8m ax=(REAL8m(1-AASIZE)/REAL8m(2*AASIZE));
	  ax<0.5;ax+=(1./REAL8m(AASIZE))) {
	for(REAL8m ay=(REAL8m(1-AASIZE)/REAL8m(2*AASIZE));
	    ay<0.5;ay+=(1./REAL8m(AASIZE))) {
	  if(poly.IsInside(PlanePoint(double(x)+ax,double(y)+ay))) {
	    incount++;
	  }
	}
      }
      if(incount>=AASIZE*AASIZE) SetPixel(x,y,color);
      else if(incount>0) {
	pixcolor.Blend(REAL8m(incount)/REAL8m(AASIZE*AASIZE),color,
		       REAL8m(AASIZE*AASIZE-incount)/REAL8m(AASIZE*AASIZE),
		       GetPixel(x,y));
	SetPixel(x,y,pixcolor);
      }
    }
#   undef AASIZE
  } else { // No antialiasing
    for(int x=axmin;x<=axmax;x++) for(int y=aymin;y<=aymax;y++) {
      if(poly.IsInside(PlanePoint(double(x),double(y)))) {
	SetPixel(x,y,color);
      }
    }
  }
}

void OommfBitmap::DrawPolyLine(const Nb_List<PlanePoint>& vlist,
			       REAL8m linewidth,OommfPackedRGB color,
			       int miter,int antialias)
{ // Draws (potentially fat) line segments between each successive
  // pair of points in vlist.  If the miter!=0, then the segment
  // ends are mitered.  In general, this is an unclosed figure,
  // unless the last and first point are equal, in which case those
  // ends will be mitered if requested.

  const PlanePoint *pta,*ptb;  // Polyline vertex point pair
  PlanePoint qa1,qa2,qb1,qb2; // Fat line corner points
  const PlanePoint *rta=NULL; // First corner point
  PlanePoint ra1,ra2;         // Corner points for 1st vertex of 1st line

  OommfPolygon fatline;

  REAL8m offset=linewidth/2.;

  Nb_List_Index<PlanePoint> key;
  pta=vlist.GetFirst(key);
  while((ptb=vlist.GetNext(key))!=NULL) {
    // Construct perpendicular to line segment
    REAL8m zy= ptb->x - pta->x;
    REAL8m zx= pta->y - ptb->y;
    REAL8m magsq=zx*zx+zy*zy;
    if(magsq>REAL8_EPSILON) {
      // Line segment big enough to work with
      REAL8m adj=offset/sqrt(magsq);
      zx*=adj; zy*=adj;
      qa1.x=pta->x+zx;      qa2.x=pta->x-zx;
      qa1.y=pta->y+zy;      qa2.y=pta->y-zy;
      if(miter && rta!=NULL) {
	// Miter ends with last segment
	fatline.Clear();
	fatline.AddVertex(qb1); fatline.AddVertex(qa1);
	fatline.AddVertex(qb2); fatline.AddVertex(qa2);
	DrawFilledPoly(fatline,color,antialias);
      }
      qb1.x=ptb->x+zx;      qb2.x=ptb->x-zx;
      qb1.y=ptb->y+zy;      qb2.y=ptb->y-zy;
      fatline.Clear();
      fatline.AddVertex(qa1);  fatline.AddVertex(qa2);
      fatline.AddVertex(qb2);  fatline.AddVertex(qb1);
      DrawFilledPoly(fatline,color,antialias);
    }
    if(rta==NULL) {
      rta=pta; ra1=qa1; ra2=qa2; // Save 1st vertex info
    }
    pta=ptb;
  }
  if(miter && rta!=NULL && rta->x==pta->x && rta->y==pta->y) {
    // Miter first and last segments
    fatline.Clear();
    fatline.AddVertex(qb1); fatline.AddVertex(ra1);
    fatline.AddVertex(qb2); fatline.AddVertex(ra2);
    DrawFilledPoly(fatline,color,antialias);
  }
}

void OommfBitmap::DrawFilledArrow(REAL8m xc,REAL8m yc,REAL8m mag,
                                 REAL8m xcos,REAL8m ycos,REAL8m zcos,
                                 OommfPackedRGB color,BOOL antialias)
{ // Draws arrow centered at point (xc,yc) in bitmap, of size mag,
  // directional cosines xcos,ycos,zcos, with specified color.
  work_poly=default_arrow;
  REAL8m magadj=1.-zcos*zcos;
  if(magadj<REAL8_EPSILON) return;  // In-plane comp. too small to display
  magadj=sqrt(magadj);
  work_poly.Scale(mag*magadj,mag);
  work_poly.Rotate(xcos/magadj,ycos/magadj);
  work_poly.Shift(xc,yc);
  DrawFilledPoly(work_poly,color,antialias);
}

void OommfBitmap::DrawDiamondWithCross(REAL8m xc,REAL8m yc,REAL8m mag,
				       OommfPackedRGB outercolor,
				       OommfPackedRGB innercolor)
{ // Symbol representing out-of-plane vectors
  xc=OC_ROUND(xc); yc=OC_ROUND(yc); // Make it look nice
  work_poly=default_diamond;
  work_poly.Scale(mag,mag);
  work_poly.AdjustAllToIntegers();
  work_poly.Shift(xc,yc);
  DrawFilledPoly(work_poly,outercolor,0);
  if(mag>MinCrossDisplayMag) {
    work_poly=default_cross;
    work_poly.Scale(mag,mag);
    work_poly.Shift(xc,yc);
    DrawFilledPoly(work_poly,innercolor,0);
  }
}

void OommfBitmap::DrawSquareWithDot(REAL8m xc,REAL8m yc,REAL8m mag,
				    OommfPackedRGB outercolor,
				    OommfPackedRGB innercolor)
{ // Symbol representing in-to-plane vectors.
  xc=OC_ROUND(xc); yc=OC_ROUND(yc); // Make it look nice
  REAL8m radius1=ceil(InOutTipRadius*mag);
  DrawFilledRectangle(xc-radius1,yc-radius1,xc+radius1,yc+radius1,
		      outercolor);
  REAL8m radius2=floor(DotRadius*mag);
  if(radius2>0.5) {
    DrawFilledRectangle(xc-radius2,yc-radius2,xc+radius2,yc+radius2,
			innercolor);
  }
}

int OommfBitmap::WritePPMFile(const char* filename,int ppm_type)
{ // Use filename == NULL or "-" to write to stdout.
  // "ppm_type" corresponds to the PPM magic number.  It should be either
  // 3 (ASCII text) or 6 (binary).
  // Return value is
  //     0 => No error
  //     1 => Bad parameter
  //     2 => Unable to open filename for output
  //     4 => Write error
#define MEMBERNAME "WritePPMFile(char*,int)"
  if(ppm_type!=3 && ppm_type!=6) {
    NonFatalError(STDDOC,"Illegal PPM format type requested: %d."
		  "  Should be either 3 or 6.",ppm_type);
    return 1;
  }

  // Open file and write header
  FILE* fptr;
  if(filename==NULL || strcmp(filename,"-")==0) fptr=stdout;
  else if((fptr=fopen(filename,"w"))==NULL) {
    NonFatalError(STDDOC,"Unable to open file %s for output",filename);
    return 2;
  }
  setvbuf(fptr,(char *)NULL,_IOFBF,8192);

#if (TKCOLORLENGTH != 6)
# error "Current OommfBitmap code only supports TKCOLORLENGTH==6"
#endif

  int errorcode=0;
  if(ppm_type==6) {
    fprintf(fptr,"P6\n%d %d\n%d\n",xsize,ysize,255);
    for(int i=0;i<totalsize;i++) {
      // Note: The following code breaks encapsulation of
      //       the OommfPackedRGB class.
      if(fwrite((char *)(bitmap+i),1,3,fptr) != 3) {
	errorcode=4;
	break;
      }
    }
  } else {
    unsigned int red,green,blue;
    fprintf(fptr,"P3\n%d %d\n%d\n",xsize,ysize,255);
    for(int i=0;i<totalsize;i++) {
      bitmap[i].Get(red,green,blue);
      fprintf(fptr,"%u %u %u\n",red,green,blue);
      if(feof(fptr) || ferror(fptr)) {
	errorcode=4;
	break;
      }
    }
  }
  fclose(fptr);
  return errorcode;
#undef MEMBERNAME

}

int OommfBitmap::WritePPMChannel(Tcl_Channel channel,int ppm_type)
{ // "ppm_type" corresponds to the PPM magic number.  It should be either
  // 3 (ASCII text) or 6 (binary).
  // Return value is
  //     0 => No error
  //     1 => Bad parameter
  //     4 => Write error
#define MEMBERNAME "WritePPMChannel(Tcl_Channel,int)"
  if(ppm_type!=3 && ppm_type!=6) {
    NonFatalError(STDDOC,"Illegal PPM format type requested: %d."
		  "  Should be either 3 or 6.",ppm_type);
    return 1;
  }

  // Write header
#if (TKCOLORLENGTH != 6)
# error "Current OommfBitmap code only supports TKCOLORLENGTH==6"
#endif

  /// Note: On some (non-ANSI) compilers, sprintf() returns a pointer
  ///       to the buffer instead of an output byte count.
  if(ppm_type==6) {
    char scratch[256];  // Make sure this is big enough as used below!
    sprintf(scratch,"P6\n%d %d\n%d\n",xsize,ysize,255);
    Tcl_Write(channel,scratch,strlen(scratch));
    for(int i=0;i<totalsize;i++) {
      // Note: The following code breaks encapsulation of
      //       the OommfPackedRGB class.
      if(Tcl_Write(channel,(char *)(bitmap+i),3) != 3) {
	return 4;
      }
    }
  } else {
    char scratch[1024];  // Make sure this is big enough as used below!
    sprintf(scratch,"P3\n%d %d\n%d\n",xsize,ysize,255);
    Tcl_Write(channel,scratch,strlen(scratch));
    unsigned int comp[12];
    int i,chunksize;
    for(i=0;i<totalsize-3;i+=4) {
      // Unrolled loop
      bitmap[i].Get(comp[0],comp[1],comp[2]);   // Red, Green, Blue
      bitmap[i+1].Get(comp[3],comp[4],comp[5]);
      bitmap[i+2].Get(comp[6],comp[7],comp[8]);
      bitmap[i+3].Get(comp[9],comp[10],comp[11]);
      sprintf(scratch,"%u %u %u\n%u %u %u\n%u %u %u\n%u %u %u\n",
	      comp[0],comp[1],comp[2],comp[3],comp[4],comp[5],
	      comp[6],comp[7],comp[8],comp[9],comp[10],comp[11]);
      chunksize=strlen(scratch);
      if(Tcl_Write(channel,scratch,chunksize) != chunksize) {
	return 4;
      }
    }
    for(;i<totalsize;i++) {
      // Remainder
      bitmap[i].Get(comp[0],comp[1],comp[2]);
      sprintf(scratch,"%u %u %u\n",comp[0],comp[1],comp[2]);
      chunksize=strlen(scratch);
      if(Tcl_Write(channel,scratch,chunksize) != chunksize) {
	return 4;
      }
    }
  }
  return 0;
#undef MEMBERNAME
}
